/* KallistiOS 0.7

   fs_dcloadconsole.c
   (c)2000 Andrew Kieschnick
   
*/

/*

This is a rewrite of Dan Potter's fs_serconsole to use the dcload / dc-tool
fileserver and console. Other than that, it acts like fs_serconsole - whatever
you export from the PC is in /pc on KOS, and kernel console messages go to the
PC.

There are a few stupid hacks in here.

*/

#include <malloc.h>
#include <kallisti/fs/dcloadconsole.h>
#include <kallisti/thread.h>
#include <kallisti/process.h>

/******************************************************** Serial interface ****/
/* Access mutex (can only do one thing with the serial port at once) */
static thd_mutex_t mutex;

/* Printk replacement */
static void (*old_printk)(char *str) = NULL;
void dcload_printk(char *str) {
    thd_mutex_lock(&mutex);
    dcloadsyscall(DCLOAD_WRITE, 1, str, strlen(str));
    thd_mutex_unlock(&mutex);
}

static char *dcload_path = NULL;
uint32 dcload_open(const char *fn, int mode) {
    int hnd = 0;
    uint32 h;
    int dcload_mode = 0;
    
    thd_mutex_lock(&mutex);
    
    if (mode & O_DIR) {
	hnd = dcloadsyscall(DCLOAD_OPENDIR, fn);
	if (hnd) {
	    if (dcload_path)
		free(dcload_path);
	    if (fn[strlen(fn)] == '/') {
		dcload_path = malloc(strlen(fn)+1);
		strcpy(dcload_path, fn);
	    } else {
		dcload_path = malloc(strlen(fn)+2);
		strcpy(dcload_path, fn);
		strcat(dcload_path, "/");
	    }
	}
    } else { /* hack */
	if (mode == O_RDONLY)
	    dcload_mode = 0;
	if (mode == O_RDWR)
	    dcload_mode = 2 | 0x0200;
	if (mode == O_WRONLY)
	    dcload_mode = 1 | 0x0200;
	if (mode == O_APPEND)
	    dcload_mode =  2 | 8 | 0x0200;
	hnd = dcloadsyscall(DCLOAD_OPEN, fn, dcload_mode, 0644);
	hnd++; /* KOS uses 0 for error, not -1 */
    }
    
    h = hnd;

    thd_mutex_unlock(&mutex);
    return h;
}

void dcload_close(uint32 hnd) {
    thd_mutex_lock(&mutex);

    if (hnd)
	if (hnd > 100) /* hack */
	    dcloadsyscall(DCLOAD_CLOSEDIR, hnd);
	else {
	    hnd--; /* KOS uses 0 for error, not -1 */
	    dcloadsyscall(DCLOAD_CLOSE, hnd);
	}
    thd_mutex_unlock(&mutex);
}

ssize_t dcload_read(uint32 hnd, void *buf, size_t cnt) {
    ssize_t ret;

    thd_mutex_lock(&mutex);
    
    if (hnd) {
	hnd--; /* KOS uses 0 for error, not -1 */
	ret = dcloadsyscall(DCLOAD_READ, hnd, buf, cnt);
    }

    thd_mutex_unlock(&mutex);
    return ret;
}

ssize_t dcload_write(uint32 hnd, const void *buf, size_t cnt) {
    ssize_t ret;
    	
    thd_mutex_lock(&mutex);
    
    if (hnd) {
	hnd--; /* KOS uses 0 for error, not -1 */
	ret = dcloadsyscall(DCLOAD_WRITE, hnd, buf, cnt);
    }

    thd_mutex_unlock(&mutex);
    return ret;
}

off_t dcload_seek(uint32 hnd, off_t offset, int whence) {
    off_t ret;

    thd_mutex_lock(&mutex);

    if (hnd) {
	hnd--; /* KOS uses 0 for error, not -1 */
	ret = dcloadsyscall(DCLOAD_LSEEK, hnd, offset, whence);
    }

    thd_mutex_unlock(&mutex);
    return ret;
}

off_t dcload_tell(uint32 hnd) {
    off_t ret;
    
    thd_mutex_lock(&mutex);

    if (hnd) {
	hnd--; /* KOS uses 0 for error, not -1 */
	ret = dcloadsyscall(DCLOAD_LSEEK, hnd, 0, SEEK_CUR);
    }

    thd_mutex_unlock(&mutex);
    return ret;
}

size_t dcload_total(uint32 hnd) {
    size_t ret;
    size_t cur;
	
    thd_mutex_lock(&mutex);
	
    if (hnd) {
	hnd--; /* KOS uses 0 for error, not -1 */
	cur = dcloadsyscall(DCLOAD_LSEEK, hnd, 0, SEEK_CUR);
	ret = dcloadsyscall(DCLOAD_LSEEK, hnd, 0, SEEK_END);
	dcloadsyscall(DCLOAD_LSEEK, hnd, cur, SEEK_SET);
    }
	
    thd_mutex_unlock(&mutex);
    return ret;
}

/* Not thread-safe, but that's ok because neither is the FS */
static dirent_t dirent;
dirent_t *dcload_readdir(uint32 hnd) {
    dirent_t *rv = NULL;
    dcload_dirent_t *dcld;
    dcload_stat_t filestat;
    char *fn;

    if (hnd < 100) return NULL; /* hack */

    thd_mutex_lock(&mutex);

    dcld = (dcload_dirent_t *)dcloadsyscall(DCLOAD_READDIR, hnd);
    
    if (dcld) {
	rv = &dirent;
	strcpy(rv->name, dcld->d_name);
	rv->size = 0;
	rv->time = 0;
	rv->attr = 0; /* what the hell is attr supposed to be anyways? */

	fn = malloc(strlen(dcload_path)+strlen(dcld->d_name)+1);
	strcpy(fn, dcload_path);
	strcat(fn, dcld->d_name);

	if (!dcloadsyscall(DCLOAD_STAT, fn, &filestat)) {
	    if (filestat.st_mode & S_IFDIR)
		rv->size = -1;
	    else
		rv->size = filestat.st_size;
	    rv->time = filestat.st_mtime;
	    
	}
	
	free(fn);
    }
    
    thd_mutex_unlock(&mutex);
    return rv;
}

int dcload_rename(const char *fn1, const char *fn2) {
    int ret;

    thd_mutex_lock(&mutex);

    /* really stupid hack, since I didn't put rename() in dcload */

    ret = dcloadsyscall(DCLOAD_LINK, fn1, fn2);

    if (!ret)
	ret = dcloadsyscall(DCLOAD_UNLINK, fn1);

    thd_mutex_unlock(&mutex);
    return ret;
}

int dcload_unlink(const char *fn) {
    int ret;

    thd_mutex_lock(&mutex);

    ret = dcloadsyscall(DCLOAD_UNLINK, fn);

    thd_mutex_unlock(&mutex);
    return ret;
}

/* Pull all that together */
static vfs_handler vh = {
	0, 0,		/* In-kernel, no cacheing */
	dcload_open,
	dcload_close,
	dcload_read,
	dcload_write,
	dcload_seek,
	dcload_tell,
	dcload_total,
	dcload_readdir,
	NULL,
	dcload_rename,
	dcload_unlink
};

unsigned char *dcload_wrkmem = 0;
/* This is called before any main() gets to take a whack */
int fs_dcloadconsole_init() {
	/* Init thread mutex */
	thd_mutex_reset(&mutex);

	/* Check for dcload */
	if (*DCLOADMAGICADDR != DCLOADMAGICVALUE)
	    return -1;

	/* Give dcload the 64k it needs to compress data */
	dcload_wrkmem = malloc(65536);
	if (dcload_wrkmem)
	    dcloadsyscall(DCLOAD_ASSIGNWRKMEM, dcload_wrkmem);

	/* Set up printk replacement */
	old_printk = ps_set_printk(dcload_printk);

	/* Register with VFS */
	return fs_handler_add("/pc", &vh);
}

int fs_dcloadconsole_shutdown() {
        dcloadsyscall(DCLOAD_EXIT);
        ps_set_printk(old_printk);
	free(dcload_wrkmem);
	return fs_handler_remove(&vh);
}

